@@ -3,13 +3,15 @@ import numbers
 from bisect import bisect_left
 from collections import UserDict, namedtuple
 from collections.abc import MutableMapping, Mapping
-from copy import deepcopy
+import copy
 from keyword import iskeyword
 from operator import attrgetter
 from typing import (Any, Callable, Dict, Generic, Hashable, Iterator,
                     List, Optional, Tuple, TypeVar, Union, Sequence,
                     Set)
 
+from pythonic_toolbox.decorators.decorator_utils import method_synchronized
+
 T = TypeVar("T")
 KT = TypeVar("KT")
 VT = TypeVar("VT")
@@ -83,7 +85,7 @@ def select_list_of_dicts(dict_lst: List[Dict],
     keys = keys or []
     keys = list(keys)  # make a shallow copy
 
-    dict_lst = deepcopy(dict_lst)
+    dict_lst = copy.deepcopy(dict_lst)
     res: Union[List[Dict], Iterator[Dict]] = dict_lst
 
     if preds:
@@ -107,7 +109,7 @@ def unique_list_of_dicts(dict_list: List[Dict]) -> List[Dict]:
     for d in dict_list:
         items_tuple = tuple(d.items())
         if items_tuple not in items_tuple_set:
-            unique_res.append(deepcopy(d))
+            unique_res.append(copy.deepcopy(d))
             items_tuple_set.add(items_tuple)
     return unique_res
 
@@ -130,7 +132,7 @@ def walk_leaves(data: Optional[Union[Dict, List]] = None,
         obj = data
     else:
         # won't touch the original data
-        obj = deepcopy(data)
+        obj = copy.deepcopy(data)
 
     if trans_fun is None:
         return obj if inplace is False else None
@@ -230,7 +232,7 @@ class _MyUserDict(MutableMapping):
 class DictObj(_MyUserDict):
     def __init__(self, in_dict: Dict):
 
-        in_dict = deepcopy(in_dict)
+        in_dict = copy.deepcopy(in_dict)
 
         if any(map(lambda key: not isinstance(key, str),
                    in_dict.keys())):
@@ -250,29 +252,35 @@ class DictObj(_MyUserDict):
         else:
             return data
 
+    @method_synchronized
     def __setitem__(self, key, item):
         self._user_dict_hidden_data[key] = self._create_obj_or_keep(item)
 
+    @method_synchronized
     def popitem(self):
         """
         Override popitem from MutableMapping, make behavior popitem FILO like ordinary dict since 3.6
         """
         return self._user_dict_hidden_data.popitem()
 
+    @method_synchronized
     def pop(self, key):
         val = self._user_dict_hidden_data[key]
         del self._user_dict_hidden_data[key]
         return val
 
+    @method_synchronized
     def __getattribute__(self, item):
-        if item == '_user_dict_hidden_data':
-            return self.__dict__['_user_dict_hidden_data']
+        if item in {'_user_dict_hidden_data'}:
+            return object.__getattribute__(self, '__dict__')[item]
         else:
             return super(DictObj, self).__getattribute__(item)
 
+    @method_synchronized
     def __delitem__(self, key):
         del self._user_dict_hidden_data[key]
 
+    @method_synchronized
     def __setattr__(self, key, value):
         """DictObj that can change attribute"""
         if key == '_user_dict_hidden_data':
@@ -283,25 +291,27 @@ class DictObj(_MyUserDict):
                 # handle case when accessing attribute directly
                 # by adding '_' for keyword/non-identifier attribute
                 key = key[1:]
+            cls_name = object.__getattribute__(self, '__class__').__name__
             data[key] = self._create_obj_or_keep(value)
             object.__setattr__(self, '_user_dict_hidden_data', data)
 
+    @method_synchronized
     def __getattr__(self, item):
+        __dict__ = object.__getattribute__(self, '__dict__')
         try:
-            res = self.__dict__['_user_dict_hidden_data'][item]
+            return __dict__['_user_dict_hidden_data'][item]
         except KeyError:
             if len(item) >= 2 and item.startswith('_') and not item.startswith('__'):
                 # keyword like attribute can be accessed by adding "_" in prefix
                 new_item = item[1:]
                 if new_item.isidentifier() is False or iskeyword(new_item):
                     try:
-                        return self.__dict__['_user_dict_hidden_data'][new_item]
+                        return __dict__['_user_dict_hidden_data'][new_item]
                     except KeyError:
                         pass
             raise AttributeError(f'AttributeError {item}')
-        else:
-            return res
 
+    @method_synchronized
     def __delattr__(self, item):
         try:
             del self[item]
@@ -326,13 +336,28 @@ class DictObj(_MyUserDict):
         """not hashable"""
         return None
 
-    def to_dict(self):
+    @method_synchronized
+    def __copy__(self):
+        my_copy = type(self)({})
+        my_copy._user_dict_hidden_data = copy.copy(self._user_dict_hidden_data)
+        return my_copy
+
+    @method_synchronized
+    def __deepcopy__(self, memo=None):
+        if memo is None:
+            memo = {}
+        my_copy = type(self)({})
+        my_copy._user_dict_hidden_data = copy.deepcopy(self._user_dict_hidden_data, memo)
+        return my_copy
+
+    @method_synchronized
+    def to_dict(self, flatten=True):
         result = {}
         for key, item in self._user_dict_hidden_data.items():
             if isinstance(item, (list, tuple)):
                 result[key] = [x.to_dict() if hasattr(x, 'to_dict') and callable(getattr(x, 'to_dict')) else x
                                for x in item]
-            elif isinstance(item, DictObj):
+            elif isinstance(item, DictObj) and flatten:
                 result[key] = item.to_dict()
             else:
                 result[key] = item
@@ -354,9 +379,7 @@ class FinalDictObj(DictObj):
     __frozen_err_msg = 'Cannot modify attribute/item in an already initialized FinalDictObj'
 
     def __init__(self, in_dict: Dict):
-
-        in_dict = deepcopy(in_dict)
-
+        in_dict = copy.deepcopy(in_dict)
         super(FinalDictObj, self).__init__(in_dict)
         self._freeze()
 
@@ -369,22 +392,27 @@ class FinalDictObj(DictObj):
         else:
             return data
 
+    @method_synchronized
     def _freeze(self):
         self.__is_frozen = True
 
+    @method_synchronized
     @_frozen_checker
     def __setitem__(self, key, value):
         """DictObj that cannot change attribute"""
         super(FinalDictObj, self).__setitem__(key, value)
 
+    @method_synchronized
     @_frozen_checker
     def __delitem__(self, key):
         super(FinalDictObj, self).__delitem__(key)
 
+    @method_synchronized
     @_frozen_checker
     def popitem(self):
         return super(FinalDictObj, self).popitem()
 
+    @method_synchronized
     @_frozen_checker
     def pop(self, key):
         return super(FinalDictObj, self).pop(key)
@@ -392,25 +420,38 @@ class FinalDictObj(DictObj):
     def __setattr__(self, key, value):
         """DictObj that cannot change attribute"""
         if key == '_FinalDictObj__is_frozen':
-            if value is True:
-                # __is_frozen can only be assigned as True
-                object.__setattr__(self, '_FinalDictObj__is_frozen', True)
-            else:
-                raise RuntimeError('__is_frozen can only be assigned as True')
+            object.__setattr__(self, '_FinalDictObj__is_frozen', value)
         else:
             if self.__is_frozen:
                 raise RuntimeError(self.__frozen_err_msg)
 
             super(FinalDictObj, self).__setattr__(key, value)
 
+    @method_synchronized
     @_frozen_checker
     def __delattr__(self, item):
         super(FinalDictObj, self).__delattr__(item)
 
+    @method_synchronized
     @_frozen_checker
     def update(self, *args, **kwargs):
         super(FinalDictObj, self).update(*args, **kwargs)
 
+    @method_synchronized
+    def __copy__(self):
+        my_copy = type(self)({})
+        my_copy._FinalDictObj__is_frozen = False
+        my_copy._user_dict_hidden_data = copy.copy(self._user_dict_hidden_data)
+        my_copy._FinalDictObj__is_frozen = True
+        return my_copy
+
+    @method_synchronized
+    def __deepcopy__(self, memo=None):
+        if memo is None:
+            memo = {}
+        my_copy = type(self)(copy.deepcopy(self.to_dict(), memo))
+        return my_copy
+
 
 class RangeKeyDict(Generic[KT, VT]):
     """
