@@ -1,10 +1,13 @@
 import json
 from collections import defaultdict
-from typing import List, Iterable, Union, Optional, Callable, TypeVar, Any, Tuple, DefaultDict, Hashable
+from functools import singledispatch
+from typing import (List, Iterable, Generator, Union, Optional, Callable,
+                    TypeVar, Any, Tuple, DefaultDict, Hashable, overload)
 
 from funcy import first, identity
 
 T = TypeVar("T")
+T1 = TypeVar("T1")
 
 
 def sort_with_custom_orders(values: List[T],
@@ -81,13 +84,13 @@ def sort_with_custom_orders(values: List[T],
 
     order_map: DefaultDict[Hashable, int] = defaultdict(lambda: 1)
     for idx, item in enumerate(prefix_orders):
-        order_map[_hash_fun(item)] = idx - len(prefix_orders)
+        order_map[_hash_fun(key(item))] = idx - len(prefix_orders)
 
     for idx, item in enumerate(suffix_orders, start=2):
-        order_map[_hash_fun(item)] = idx
+        order_map[_hash_fun(key(item))] = idx
 
     def key_func(x: T) -> Tuple[int, Any]:
-        return order_map[hash_fun(x)], key(x)
+        return order_map[hash_fun(key(x))], key(x)
 
     sorted_values = sorted(values, key=key_func, reverse=reverse)
 
@@ -122,14 +125,45 @@ def until(values: Optional[Union[List[T], Iterable]],
         raise ValueError('values type should be list, Iterable')
 
 
+@overload
 def unpack_list(source: List[Any], target_num: int, default: Optional[Any] = None) -> List[Any]:
+    pass
+
+
+@overload
+def unpack_list(source: Union[Generator, Iterable, range], target_num: int, default: Optional[Any] = None) -> List[Any]:
+    pass
+
+
+@singledispatch
+def unpack_list(source: Union[Generator, Iterable, range], target_num: int, default: Optional[Any] = None) -> List[Any]:
+    source_iter = iter(source)
+    res = []
+    cnt = 0
+    try:
+        while cnt < target_num:
+            tmp = next(source_iter)
+            res.append(tmp)
+            cnt += 1
+    except StopIteration:
+        pass
+
+    while cnt < target_num:
+        res.append(default)
+        cnt += 1
+
+    return res
+
+
+@unpack_list.register(list)
+def _(source: List[Any], target_num: int, default: Optional[Any] = None) -> List[Any]:
     return [*source, *([default] * (target_num - len(source)))] if len(source) < target_num else source[:target_num]
 
 
-def filter_allowable(candidates: Optional[List[Any]] = None,
-                     allow_list: Optional[List[T]] = None,
-                     block_list: [List[T]] = None,
-                     key: Optional[Callable[..., Any]] = None) -> Iterable[Any]:
+def filter_allowable(candidates: Optional[List[T]] = None,
+                     allow_list: Optional[List[T1]] = None,
+                     block_list: Optional[List[T1]] = None,
+                     key: Optional[Callable[..., T1]] = None) -> Iterable[T]:
     if key is None:
         key = identity
     allow_list = allow_list or set()
